@@ -22,22 +22,29 @@
 #include <signal.h>
 #include <sys/wait.h>
 #include <X11/Xauth.h>
+#include <sys/types.h>
+#include <fcntl.h>
 
 #include "gdm.h"
+#include "gdm-net.h"
 #include "server.h"
 #include "display.h"
 #include "slave.h"
 #include "misc.h"
 #include "xdmcp.h"
 #include "choose.h"
+#include "gdm-net.h"
 
-static const gchar RCSid[]="$Id: display.c,v 1.14 2001/07/05 10:27:29 jirka Exp $";
+static const gchar RCSid[]="$Id: display.c,v 1.23 2001/10/26 03:43:28 jirka Exp $";
 
 /* External vars */
 extern gboolean GdmXdmcp;
 extern gint sessions;
+extern gint flexi_servers;
 extern gint pending;
 extern GSList *displays;
+extern GdmConnection *fifoconn;
+extern GdmConnection *unixconn;
 
 static gboolean
 gdm_display_check_loop (GdmDisplay *disp)
@@ -84,10 +91,7 @@ gdm_display_check_loop (GdmDisplay *disp)
    */
   if (disp->retry_count > 4)
     {
-      gchar *msg;
-      msg = g_strdup_printf (_("Failed to start the display server several times in a short time period; disabling display %s"), disp->name);
-      gdm_error (msg);
-      g_free (msg);
+      gdm_error (_("Failed to start the display server several times in a short time period; disabling display %s"), disp->name);
       disp->disabled = TRUE;
 
       gdm_debug ("Failed to start X server after several retries; aborting.");
@@ -128,7 +132,8 @@ gdm_display_check_loop (GdmDisplay *disp)
 gboolean 
 gdm_display_manage (GdmDisplay *d)
 {
-    sigset_t mask, omask;
+    pid_t pid;
+    int i;
 
     if (!d) 
 	return FALSE;
@@ -138,21 +143,23 @@ gdm_display_manage (GdmDisplay *d)
     if ( ! gdm_display_check_loop (d))
 	    return FALSE;
 
+    gdm_sigchld_block_push ();
+
     /* If we have an old slave process hanging around, kill it */
     if (d->slavepid) {
-	sigemptyset (&mask);
-	sigaddset (&mask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &mask, &omask); 
-
-	if (kill (d->slavepid, SIGINT) == 0)
-		waitpid (d->slavepid, 0, 0);
-	d->slavepid = 0;
-
-	sigprocmask (SIG_SETMASK, &omask, NULL);
+	    if (kill (d->slavepid, SIGINT) == 0)
+		    waitpid (d->slavepid, 0, 0);
+	    d->slavepid = 0;
     }
 
     /* Fork slave process */
-    switch (d->slavepid = fork()) {
+    gdm_sigterm_block_push ();
+    pid = d->slavepid = fork ();
+    gdm_sigterm_block_push ();
+
+    gdm_sigchld_block_pop ();
+
+    switch (pid) {
 
     case 0:
 	setpgid (0, 0);
@@ -161,11 +168,31 @@ gdm_display_manage (GdmDisplay *d)
 	if (GdmXdmcp)
 	    gdm_xdmcp_close ();
 
+	gdm_connection_close (fifoconn);
+	fifoconn = NULL;
+	gdm_connection_close (unixconn);
+	unixconn = NULL;
+
+	/* Close everything */
+	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
+	    close(i);
+
+	/* No error checking here - if it's messed the best response
+         * is to ignore & try to continue */
+	open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+	open("/dev/null", O_RDWR); /* open stdout - fd 1 */
+	open("/dev/null", O_RDWR); /* open stderr - fd 2 */
+
 	if (d->type == TYPE_LOCAL) {
 	    gdm_slave_start (d);
 	    /* if we ever return, bad things are happening */
 	    gdm_server_stop (d);
 	    _exit (DISPLAY_ABORT);
+	} else if (SERVER_IS_FLEXI (d)) {
+	    gdm_slave_start (d);
+	    gdm_server_stop (d);
+	    /* we expect to return after the session finishes */
+	    _exit (DISPLAY_REMANAGE);
 	} else if (d->type == TYPE_XDMCP && d->dispstat == XDMCP_MANAGED) {
 	    gdm_slave_start (d);
 	    gdm_server_stop (d);
@@ -181,16 +208,17 @@ gdm_display_manage (GdmDisplay *d)
 	break;
 
     case -1:
-	gdm_error (_("gdm_display_manage: Failed forking gdm slave process for %d"), d->name);
+	gdm_error (_("gdm_display_manage: Failed forking gdm slave process for %s"), d->name);
 
 	return FALSE;
 
     default:
-	gdm_debug ("gdm_display_manage: Forked slave: %d", d->slavepid);
+	gdm_debug ("gdm_display_manage: Forked slave: %d",
+		   (int)pid);
 	break;
     }
 
-    if (d->type == TYPE_LOCAL) {
+    if (SERVER_IS_LOCAL (d)) {
 	    d->dispstat = DISPLAY_SUCCESS;
     }
 
@@ -226,7 +254,7 @@ gdm_display_unmanage (GdmDisplay *d)
     
     if (d->type == TYPE_LOCAL)
 	d->dispstat = DISPLAY_DEAD;
-    else /* TYPE_XDMCP */
+    else /* TYPE_XDMCP,TYPE_FLEXI,TYPE_FLEXI_XNEST */
 	gdm_display_dispose (d);
 }
 
@@ -244,8 +272,18 @@ gdm_display_dispose (GdmDisplay *d)
     if (d == NULL)
 	return;
 
+    if (d->socket_conn != NULL) {
+	    GdmConnection *conn = d->socket_conn;
+	    d->socket_conn = NULL;
+	    gdm_connection_set_close_notify (conn, NULL, NULL);
+    }
+
+    displays = g_slist_remove (displays, d);
+
+    if (SERVER_IS_FLEXI (d))
+	    flexi_servers --;
+
     if (d->type == TYPE_XDMCP) {
-	displays = g_slist_remove (displays, d);
 	sessions--;
 	d->type = -1;
 
@@ -265,11 +303,15 @@ gdm_display_dispose (GdmDisplay *d)
     g_free (d->authfile);
     d->authfile = NULL;
 
+    g_free (d->authfile_gdm);
+    d->authfile_gdm = NULL;
+
     if (d->auths) {
 	GSList *tmpauth = d->auths;
 
 	while (tmpauth && tmpauth->data) {
 	    XauDisposeAuth ((Xauth *) tmpauth->data);
+	    tmpauth->data = NULL;
 	    tmpauth = tmpauth->next;
 	}
 
@@ -293,6 +335,15 @@ gdm_display_dispose (GdmDisplay *d)
 	    gdm_choose_indirect_dispose_empty_id (d->indirect_id);
     d->indirect_id = 0;
 
+    g_free (d->xnest_disp);
+    d->xnest_disp = NULL;
+
+    g_free (d->xnest_auth_file);
+    d->xnest_auth_file = NULL;
+
+    g_free (d->login);
+    d->login = NULL;
+
     g_free (d);
 }
 
