@@ -22,21 +22,23 @@
 #include <pwd.h>
 
 #ifdef HAVE_CRYPT
-  #include <crypt.h>
+#  include <crypt.h>
 #endif /* HAVE_CRYPT */
 
+#include <vicious.h>
+
 #include "gdm.h"
 #include "misc.h"
 #include "slave.h"
 #include "verify.h"
 
-static const gchar RCSid[]="$Id: verify-crypt.c,v 1.9 2001/04/30 23:48:11 jirka Exp $";
+static const gchar RCSid[]="$Id: verify-crypt.c,v 1.20 2001/10/28 05:00:22 jirka Exp $";
 
 
 /* Configuration option variables */
-extern gboolean GdmVerboseAuth;
 extern gboolean GdmAllowRoot;
 extern gboolean GdmAllowRemoteRoot;
+extern gint GdmRetryDelay;
 
 
 /**
@@ -52,11 +54,12 @@ extern gboolean GdmAllowRemoteRoot;
  */
 
 gchar *
-gdm_verify_user (const char *username,
+gdm_verify_user (GdmDisplay *d,
+		 const char *username,
 		 const gchar *display,
 		 gboolean local) 
 {
-    gchar *login, *passwd, *ppasswd;
+    gchar *login, *passwd, *ppasswd, *auth_errmsg;
     struct passwd *pwent;
 
     if (local)
@@ -64,7 +67,8 @@ gdm_verify_user (const char *username,
 
     /* Ask for the user's login */
     if (username == NULL) {
-	    login = gdm_slave_greeter_ctl (GDM_LOGIN, _("Login:"));
+	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, _("Please enter your username"));
+	    login = gdm_slave_greeter_ctl (GDM_LOGIN, _("Username:"));
 	    if (login == NULL ||
 		gdm_slave_greeter_check_interruption (login)) {
 		    if (local)
@@ -78,26 +82,66 @@ gdm_verify_user (const char *username,
 
     pwent = getpwnam (login);
         
-    ppasswd = !pwent ? NULL : pwent->pw_passwd;
+    ppasswd = (pwent == NULL) ? NULL : g_strdup (pwent->pw_passwd);
     
     /* Request the user's password */
-    passwd = gdm_slave_greeter_ctl (GDM_NOECHO, _("Password: "));
-    if (passwd == NULL)
+    if (pwent != NULL &&
+        ve_string_empty (ppasswd)) {
+	    /* eeek a passwordless account */
 	    passwd = g_strdup ("");
-    if (gdm_slave_greeter_check_interruption (passwd)) {
-	    if (local)
-		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
-	    g_free (login);
-	    g_free (passwd);
-	    return NULL;
+    } else {
+	    passwd = gdm_slave_greeter_ctl (GDM_NOECHO, _("Password: "));
+	    if (passwd == NULL)
+		    passwd = g_strdup ("");
+	    if (gdm_slave_greeter_check_interruption (passwd)) {
+		    if (local)
+			    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
+		    g_free (login);
+		    g_free (passwd);
+		    g_free (ppasswd);
+		    return NULL;
+	    }
     }
 
     if (local)
 	    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 
     if (pwent == NULL) {
-	    gdm_error (_("Couldn't authenticate %s"), login);
-	    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR, _("Login incorrect"));
+	    sleep (GdmRetryDelay);
+	    gdm_error (_("Couldn't authenticate user"));
+	    /* FIXME: Hmm, how are we sure that the login is username
+	     * and password.  That is the most common case but not
+	     * neccessairly true, this message needs to be changed
+	     * to allow for such cases */
+	    auth_errmsg = g_strdup_printf
+		    (_("\nIncorrect username or password.  "
+		       "Letters must be typed in the correct case.  "  
+		       "Please be sure the Caps Lock key is not enabled"));
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
+	    g_free (auth_errmsg);
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
+	    return NULL;
+    }
+
+    /* Check whether password is valid */
+    if (ppasswd == NULL || (ppasswd[0] != '\0' &&
+			    strcmp (crypt (passwd, ppasswd), ppasswd) != 0)) {
+	    sleep (GdmRetryDelay);
+	    /* FIXME: Hmm, how are we sure that the login is username
+	     * and password.  That is the most common case but not
+	     * neccessairly true, this message needs to be changed
+	     * to allow for such cases */
+	    auth_errmsg = g_strdup_printf
+		    (_("\nIncorrect username or password.  "
+		       "Letters must be typed in the correct case.  "  
+		       "Please be sure the Caps Lock key is not enabled"));
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
+	    g_free (auth_errmsg);
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }
 
@@ -105,35 +149,38 @@ gdm_verify_user (const char *username,
 	  ( ! GdmAllowRemoteRoot && ! local) ) &&
 	pwent->pw_uid == 0) {
 	    gdm_error (_("Root login disallowed on display '%s'"), display);
-	    if (GdmVerboseAuth) {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Root login disallowed"));
-	    } else {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Login incorrect"));
-	    }
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+					  _("The system administrator "
+					    "is not allowed to login "
+					    "from this screen"));
+	    /*gdm_slave_greeter_ctl_no_ret (GDM_ERRDLG,
+	      _("Root login disallowed"));*/
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }
 
     /* check for the standard method of disallowing users */
     if (pwent->pw_shell != NULL &&
-	strcmp (pwent->pw_shell, "/bin/false") == 0) {
+	(strcmp (pwent->pw_shell, "/sbin/nologin") == 0 ||
+	 strcmp (pwent->pw_shell, "/bin/true") == 0 ||
+	 strcmp (pwent->pw_shell, "/bin/false") == 0)) {
 	    gdm_error (_("User %s not allowed to log in"), login);
-	    if (GdmVerboseAuth) {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Login disabled"));
-	    } else {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Login incorrect"));
-	    }
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+					  _("\nThe system administrator"
+					    " has disabled your "
+					    "account."));
+	    /*gdm_slave_greeter_ctl_no_ret (GDM_ERRDLG,
+	      _("Login disabled"));*/
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }	
 
-    /* Check whether password is valid */
-    if (!passwd || !ppasswd || strcmp (crypt (passwd, ppasswd), ppasswd)) {
-	    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR, _("Login incorrect"));
-	    return NULL;
-    }
+    g_free (passwd);
+    g_free (ppasswd);
     
     return login;
 }
@@ -148,7 +195,8 @@ gdm_verify_user (const char *username,
  */
 
 void
-gdm_verify_setup_user (const gchar *login, const gchar *display) 
+gdm_verify_setup_user (GdmDisplay *d,
+		       const gchar *login, const gchar *display) 
 {
 }
 
@@ -156,10 +204,11 @@ gdm_verify_setup_user (const gchar *login, const gchar *display)
 /**
  * gdm_verify_cleanup:
  *
- * Unregister the user's session */
+ * Unregister the user's session
+ */
 
 void
-gdm_verify_check (void)
+gdm_verify_cleanup (GdmDisplay *d)
 {
 }
 
@@ -177,9 +226,10 @@ gdm_verify_check (void)
 }
 
 /* used in pam */
-void
-gdm_verify_env_setup (void)
+gboolean
+gdm_verify_open_session (GdmDisplay *d)
 {
+	return TRUE;
 }
 
 /* EOF */
