@@ -23,21 +23,23 @@
 #include <shadow.h>
 
 #ifdef HAVE_CRYPT
-  #include <crypt.h>
+#  include <crypt.h>
 #endif /* HAVE_CRYPT */
 
+#include <vicious.h>
+
 #include "gdm.h"
 #include "misc.h"
 #include "slave.h"
 #include "verify.h"
 
-static const gchar RCSid[]="$Id: verify-shadow.c,v 1.11 2001/05/13 06:44:29 jirka Exp $";
+static const gchar RCSid[]="$Id: verify-shadow.c,v 1.21 2001/10/28 05:00:22 jirka Exp $";
 
 
 /* Configuration option variables */
-extern gboolean GdmVerboseAuth;
 extern gboolean GdmAllowRoot;
 extern gboolean GdmAllowRemoteRoot;
+extern gint GdmRetryDelay;
 
 /**
  * gdm_verify_user:
@@ -52,9 +54,9 @@ extern gboolean GdmAllowRemoteRoot;
  */
 
 gchar *
-gdm_verify_user (const char *username, const gchar *display, gboolean local) 
+gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboolean local) 
 {
-    gchar *login, *passwd, *ppasswd;
+    gchar *login, *passwd, *ppasswd, *auth_errmsg;
     struct passwd *pwent;
     struct spwd *sp;
 
@@ -63,7 +65,8 @@ gdm_verify_user (const char *username, const gchar *display, gboolean local)
 
     if (username == NULL) {
 	    /* Ask for the user's login */
-	    login = gdm_slave_greeter_ctl (GDM_LOGIN, _("Login:"));
+	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, _("Please enter your username"));
+	    login = gdm_slave_greeter_ctl (GDM_LOGIN, _("Username:"));
 	    if (login == NULL ||
 		gdm_slave_greeter_check_interruption (login)) {
 		    if (local)
@@ -77,20 +80,21 @@ gdm_verify_user (const char *username, const gchar *display, gboolean local)
 
     pwent = getpwnam (login);
 
-    ppasswd = !pwent ? NULL : pwent->pw_passwd;
-    
+    setspent ();
+
     /* Lookup shadow password */
     sp = getspnam (login);
     
     if (sp)
-	ppasswd = sp->sp_pwdp;
+	ppasswd = g_strdup (sp->sp_pwdp);
     else
 	ppasswd = NULL;
     
-    endspent();
+    endspent ();
 
     /* Request the user's password */
-    if (gdm_string_empty (ppasswd)) {
+    if (pwent != NULL &&
+        ve_string_empty (ppasswd)) {
 	    /* eeek a passwordless account */
 	    passwd = g_strdup ("");
     } else {
@@ -102,6 +106,7 @@ gdm_verify_user (const char *username, const gchar *display, gboolean local)
 			    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 		    g_free (login);
 		    g_free (passwd);
+		    g_free (ppasswd);
 		    return NULL;
 	    }
     }
@@ -109,18 +114,42 @@ gdm_verify_user (const char *username, const gchar *display, gboolean local)
     if (local)
 	    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 
-    if (!pwent) {
-	    gdm_error (_("Couldn't authenticate %s"), login);
-	    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR, _("Login incorrect"));
+    if (pwent == NULL) {
+	    sleep (GdmRetryDelay);
+	    gdm_error (_("Couldn't authenticate user"));
+	    /* FIXME: Hmm, how are we sure that the login is username
+	     * and password.  That is the most common case but not
+	     * neccessairly true, this message needs to be changed
+	     * to allow for such cases */
+	    auth_errmsg = g_strdup_printf
+		    (_("\nIncorrect username or password.  "
+		       "Letters must be typed in the correct case.  "  
+		       "Please be sure the Caps Lock key is not enabled"));
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
+	    g_free (auth_errmsg);
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }
 
     /* Check whether password is valid */
     if (ppasswd == NULL || (ppasswd[0] != '\0' &&
 			    strcmp (crypt (passwd, ppasswd), ppasswd) != 0)) {
-	    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR, _("Login incorrect"));
+	    sleep (GdmRetryDelay);
+	    /* FIXME: Hmm, how are we sure that the login is username
+	     * and password.  That is the most common case but not
+	     * neccessairly true, this message needs to be changed
+	     * to allow for such cases */
+	    auth_errmsg = g_strdup_printf
+		    (_("\nIncorrect username or password.  "
+		       "Letters must be typed in the correct case.  "  
+		       "Please be sure the Caps Lock key is not enabled"));
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
+	    g_free (auth_errmsg);
 	    g_free (login);
 	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }
 
@@ -128,34 +157,37 @@ gdm_verify_user (const char *username, const gchar *display, gboolean local)
 	  ( ! GdmAllowRemoteRoot && ! local) ) &&
 	pwent->pw_uid == 0) {
 	    gdm_error (_("Root login disallowed on display '%s'"), display);
-	    if (GdmVerboseAuth) {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Root login disallowed"));
-	    } else {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Login incorrect"));
-	    }
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+					  _("The system administrator "
+					    "is not allowed to login "
+					    "from this screen"));
+	    /*gdm_slave_greeter_ctl_no_ret (GDM_ERRDLG,
+	      _("Root login disallowed"));*/
 	    g_free (login);
 	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }
 
     /* check for the standard method of disallowing users */
     if (pwent->pw_shell != NULL &&
-	strcmp (pwent->pw_shell, "/bin/false") == 0) {
+	(strcmp (pwent->pw_shell, "/sbin/nologin") == 0 ||
+	 strcmp (pwent->pw_shell, "/bin/true") == 0 ||
+	 strcmp (pwent->pw_shell, "/bin/false") == 0)) {
 	    gdm_error (_("User %s not allowed to log in"), login);
-	    if (GdmVerboseAuth) {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Login disabled"));
-	    } else {
-		    gdm_slave_greeter_ctl_no_ret (GDM_MSGERR,
-						  _("Login incorrect"));
-	    }
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+					  _("\nThe system administrator"
+					    " has disabled your "
+					    "account."));
+	    /*gdm_slave_greeter_ctl_no_ret (GDM_ERRDLG,
+	      _("Login disabled"));*/
 	    g_free (login);
 	    g_free (passwd);
+	    g_free (ppasswd);
 	    return NULL;
     }	
     g_free (passwd);
+    g_free (ppasswd);
 
     return login;
 }
@@ -170,7 +202,7 @@ gdm_verify_user (const char *username, const gchar *display, gboolean local)
  */
 
 void
-gdm_verify_setup_user (const gchar *login, const gchar *display) 
+gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display) 
 {
 }
 
@@ -181,7 +213,7 @@ gdm_verify_setup_user (const gchar *login, const gchar *display)
  * Unregister the user's session */
 
 void
-gdm_verify_cleanup (void)
+gdm_verify_cleanup (GdmDisplay *d)
 {
 }
 
@@ -200,9 +232,10 @@ gdm_verify_check (void)
 }
 
 /* used in pam */
-void
-gdm_verify_env_setup (void)
+gboolean
+gdm_verify_open_session (GdmDisplay *d)
 {
+	return TRUE;
 }
 
 /* EOF */
