@@ -35,6 +35,10 @@ static Language languages [] = {
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Catalan"), "ca_ES", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Chinese (simplified)"), "zh_CN", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Chinese (traditional)"), "zh_TW", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Croatian"), "hr_HR", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Czech"), "cs_CZ", 0 },
@@ -53,6 +57,8 @@ static Language languages [] = {
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|French"), "fr_FR", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Galician"), "gl_ES", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|German"), "de_DE", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Greek"), "el_GR", 0 },
@@ -79,6 +85,8 @@ static Language languages [] = {
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Portuguese"), "pt_PT", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Brazilian Portuguese"), "pt_BR", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Romanian"), "ro_RO", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Russian"), "ru_RU", 0 },
@@ -92,6 +100,8 @@ static Language languages [] = {
 	{ N_("N-Z|Swedish"), "sv_SE", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Turkish"), "tr_TR", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Ukrainian"), "uk_UA", 0 },
 	/* This is the POSIX/C locale for english, should really be in Other */
 	{ N_("Other|POSIX/C English"), "C", 0 },
 	{ NULL, NULL }
