@@ -23,14 +23,16 @@
 #include <gdk/gdkx.h>
 #include <X11/X.h>
 #include <X11/Xlib.h>
+#ifdef HAVE_LIBXINERAMA
+#include <X11/extensions/Xinerama.h>
+#endif
 #include <pwd.h>
+#include <unistd.h>
+#include <syslog.h>
 #include "gdmwm.h"
 #include "gdm.h"
 
-static const gchar RCSid[]="$Id: gdmwm.c,v 1.3 2001/05/24 22:32:02 jirka Exp $";
-
-/* This is defined and set up in gdmlogin.c */
-extern GdkRectangle screen;
+static const gchar RCSid[]="$Id: gdmwm.c,v 1.8 2001/10/20 20:45:11 jirka Exp $";
 
 typedef struct _GdmWindow GdmWindow;
 struct _GdmWindow {
@@ -38,6 +40,9 @@ struct _GdmWindow {
 	Window win;
 	Window deco;
 	Window shadow;
+	gboolean ignore_size_hints; /* for gdm windows */
+	gboolean center; /* do centering */
+	gboolean recenter; /* do re-centering */
 };
 
 static GList *windows = NULL;
@@ -54,6 +59,123 @@ static gulong XA_COMPOUND_TEXT = 0;
 
 static int trap_depth = 0;
 
+GdkRectangle *gdm_wm_allscreens = NULL;
+int gdm_wm_screens = 0;
+GdkRectangle gdm_wm_screen = {0,0,0,0};
+
+void 
+gdm_wm_screen_init (int cur_screen_num)
+{
+#ifdef HAVE_LIBXINERAMA
+	gboolean have_xinerama = FALSE;
+
+	gdk_flush ();
+	gdk_error_trap_push ();
+	have_xinerama = XineramaIsActive (GDK_DISPLAY ());
+	gdk_flush ();
+	if (gdk_error_trap_pop () != 0)
+		have_xinerama = FALSE;
+
+	if (have_xinerama) {
+		int screen_num, i;
+		XineramaScreenInfo *xscreens =
+			XineramaQueryScreens (GDK_DISPLAY (),
+					      &screen_num);
+
+
+		if (screen_num <= 0) {
+			/* should NEVER EVER happen */
+			syslog (LOG_ERR, "Xinerama active, but <= 0 screens?");
+			gdm_wm_screen.x = 0;
+			gdm_wm_screen.y = 0;
+			gdm_wm_screen.width = gdk_screen_width ();
+			gdm_wm_screen.height = gdk_screen_height ();
+
+			gdm_wm_allscreens = g_new0 (GdkRectangle, 1);
+			gdm_wm_allscreens[0] = gdm_wm_screen;
+			gdm_wm_screens = 1;
+			return;
+		}
+
+		if (screen_num <= cur_screen_num)
+			cur_screen_num = 0;
+
+		gdm_wm_allscreens = g_new0 (GdkRectangle, screen_num);
+		gdm_wm_screens = screen_num;
+
+		for (i = 0; i < screen_num; i++) {
+			gdm_wm_allscreens[i].x = xscreens[i].x_org;
+			gdm_wm_allscreens[i].y = xscreens[i].y_org;
+			gdm_wm_allscreens[i].width = xscreens[i].width;
+			gdm_wm_allscreens[i].height = xscreens[i].height;
+
+			if (cur_screen_num == i)
+				gdm_wm_screen = gdm_wm_allscreens[i];
+		}
+
+		XFree (xscreens);
+	} else
+#endif
+	{
+		gdm_wm_screen.x = 0;
+		gdm_wm_screen.y = 0;
+		gdm_wm_screen.width = gdk_screen_width ();
+		gdm_wm_screen.height = gdk_screen_height ();
+
+		gdm_wm_allscreens = g_new0 (GdkRectangle, 1);
+		gdm_wm_allscreens[0] = gdm_wm_screen;
+		gdm_wm_screens = 1;
+	}
+#if 0
+	/* for testing Xinerama support on non-xinerama setups */
+	{
+		gdm_wm_screen.x = 100;
+		gdm_wm_screen.y = 100;
+		gdm_wm_screen.width = gdk_screen_width () / 2 - 100;
+		gdm_wm_screen.height = gdk_screen_height () / 2 - 100;
+
+		gdm_wm_allscreens = g_new0 (GdkRectangle, 2);
+		gdm_wm_allscreens[0] = gdm_wm_screen;
+		gdm_wm_allscreens[1].x = gdk_screen_width () / 2;
+		gdm_wm_allscreens[1].y = gdk_screen_height () / 2;
+		gdm_wm_allscreens[1].width = gdk_screen_width () / 2;
+		gdm_wm_allscreens[1].height = gdk_screen_height () / 2;
+		gdm_wm_screens = 2;
+	}
+#endif
+}
+
+void 
+gdm_wm_set_screen (int cur_screen_num)
+{
+	if (cur_screen_num >= gdm_wm_screens || cur_screen_num < 0)
+		cur_screen_num = 0;
+
+	gdm_wm_screen = gdm_wm_allscreens[cur_screen_num];
+}
+
+/* Not really a WM function, center a gtk window by setting uposition */
+void
+gdm_wm_center_window (GtkWindow *cw) 
+{
+	GtkRequisition req;
+        gint x, y;
+
+	gtk_widget_size_request (GTK_WIDGET (cw), &req);
+
+	x = gdm_wm_screen.x + (gdm_wm_screen.width - req.width)/2;
+	y = gdm_wm_screen.y + (gdm_wm_screen.height - req.height)/2;	
+
+	if (x < gdm_wm_screen.x)
+		x = gdm_wm_screen.x;
+	if (y < gdm_wm_screen.y)
+		y = gdm_wm_screen.y;
+
+ 	gtk_widget_set_uposition (GTK_WIDGET (cw), x, y);	
+}
+
+
+
 static void
 trap_push (void)
 {
@@ -303,7 +425,7 @@ find_window (Window w, gboolean deco_ok)
 }
 
 static void
-center_x_window (Window w, Window hintwin)
+center_x_window (GdmWindow *gw, Window w, Window hintwin)
 {
 	XSizeHints hints;
 	Status status;
@@ -325,8 +447,19 @@ center_x_window (Window w, Window hintwin)
 		return;
 	}
 
+	/* allow resizing when PSize is given, just don't allow centering when
+	 * PPosition is goven */
 	can_resize = ! (hints.flags & USSize);
-	can_reposition = ! (hints.flags & USPosition);
+	can_reposition = ! (hints.flags & USPosition ||
+			    hints.flags & PPosition);
+
+	if (can_reposition && ! gw->center)
+		can_reposition = FALSE;
+
+	if (gw->ignore_size_hints) {
+		can_resize = TRUE;
+		can_reposition = TRUE;
+	}
 
 	if ( ! can_resize &&
 	     ! can_reposition) {
@@ -340,28 +473,88 @@ center_x_window (Window w, Window hintwin)
 	/* we replace the x,y and width,height with some new values */
 
 	if (can_resize) {
-		if (width > screen.width)
-			width = screen.width;
-		if (height > screen.height)
-			height = screen.height;
+		if (width > gdm_wm_screen.width)
+			width = gdm_wm_screen.width;
+		if (height > gdm_wm_screen.height)
+			height = gdm_wm_screen.height;
 	}
 
 	if (can_reposition) {
 		/* we wipe the X with some new values */
-		x = screen.x + (screen.width - width)/2;
-		y = screen.y + (screen.height - height)/2;	
+		x = gdm_wm_screen.x + (gdm_wm_screen.width - width)/2;
+		y = gdm_wm_screen.y + (gdm_wm_screen.height - height)/2;	
 
-		if (x < screen.x)
-			x = screen.x;
-		if (y < screen.y)
-			y = screen.y;
+		if (x < gdm_wm_screen.x)
+			x = gdm_wm_screen.x;
+		if (y < gdm_wm_screen.y)
+			y = gdm_wm_screen.y;
 	}
 	
 	XMoveResizeWindow (wm_disp, w, x, y, width, height);
 
+	if (gw->center && ! gw->recenter) {
+		gw->center = FALSE;
+	}
+
 	trap_pop ();
 }
 
+#ifndef MWMUTIL_H_INCLUDED
+
+typedef struct {
+    unsigned long flags;
+    unsigned long functions;
+    unsigned long decorations;
+    long input_mode;
+    unsigned long status;
+} MotifWmHints, MwmHints;
+
+#define MWM_HINTS_DECORATIONS   (1L << 1)
+
+#define MWM_DECOR_BORDER        (1L << 1)
+
+#endif /* MWMUTIL_H_INCLUDED */
+
+static gboolean
+has_deco (Window win)
+{
+	static Atom hints_atom = None;
+	MotifWmHints *hints;
+	Atom type;
+	gint format;
+	gulong nitems;
+	gulong bytes_after;
+	gboolean border = TRUE;
+
+	trap_push ();
+
+	if (hints_atom == None)
+		hints_atom = XInternAtom (wm_disp, "_MOTIF_WM_HINTS", FALSE);
+
+	hints = NULL;
+
+	XGetWindowProperty (wm_disp, win,
+			    hints_atom, 0,
+			    sizeof (MotifWmHints) / sizeof (long),
+			    False, AnyPropertyType, &type, &format, &nitems,
+			    &bytes_after, (guchar **)&hints);
+
+	if (type != None &&
+	    hints != NULL &&
+	    hints->flags & MWM_HINTS_DECORATIONS &&
+	    ! (hints->decorations & MWM_DECOR_BORDER)) {
+		border = FALSE;
+	}
+
+	if (hints != NULL)
+		XFree (hints);
+
+	trap_pop ();
+
+	return border;
+}
+
+
 static void
 add_deco (GdmWindow *w)
 {
@@ -373,6 +566,11 @@ add_deco (GdmWindow *w)
 
 	trap_push ();
 
+	if ( ! has_deco (w->win)) {
+		trap_pop ();
+		return;
+	}
+
 	XGetGeometry (wm_disp, w->win,
 		      &root, &x, &y, &width, &height, &border, &depth);
 
@@ -399,8 +597,6 @@ add_deco (GdmWindow *w)
 				       0, 
 				       black, black);
 
-	XMapWindow (wm_disp, w->deco);
-
 	XGetWindowAttributes (wm_disp, w->deco, &attribs);
 	XSelectInput (wm_disp, w->deco,
 		      attribs.your_event_mask |
@@ -408,18 +604,37 @@ add_deco (GdmWindow *w)
 		      SubstructureNotifyMask |
 		      SubstructureRedirectMask);
 
+	XMapWindow (wm_disp, w->deco);
+
 	XReparentWindow (wm_disp, w->win, w->deco, 1, 1);
 
 	trap_pop ();
 }
 
+static gboolean
+is_wm_class (XClassHint *hint, const char *string, int len)
+{
+	if (len > 0) {
+		return ((hint->res_name != NULL &&
+			 strncmp (hint->res_name, string, len) == 0) ||
+			(hint->res_class != NULL &&
+			 strncmp (hint->res_class, string, len) == 0));
+	} else {
+		return ((hint->res_name != NULL &&
+			 strcmp (hint->res_name, string) == 0) ||
+			(hint->res_class != NULL &&
+			 strcmp (hint->res_class, string) == 0));
+	}
+}
+
 static GdmWindow *
-add_window (Window w)
+add_window (Window w, gboolean center)
 {
 	GdmWindow *gw;
 
 	gw = find_window (w, FALSE);
 	if (gw == NULL) {
+		XClassHint hint = { NULL, NULL };
 		int x, y;
 		Window root;
 		unsigned int width, height, border, depth;
@@ -430,6 +645,11 @@ add_window (Window w)
 
 		trap_push ();
 
+		/* add "centering" */
+		gw->ignore_size_hints = FALSE;
+		gw->center = center;
+		gw->recenter = FALSE;
+
 		/* hack, set USpos/size on login window */
 		if (w == wm_login_window) {
 			long ret;
@@ -437,6 +657,24 @@ add_window (Window w)
 			XGetWMNormalHints (wm_disp, w, &hints, &ret);
 			hints.flags |= USPosition | USSize;
 			XSetWMNormalHints (wm_disp, w, &hints);
+			gw->center = FALSE;
+			gw->recenter = FALSE;
+		} else if (XGetClassHint (wm_disp, w, &hint)) {
+			if (is_wm_class (&hint, "gdm", 3)) {
+				gw->ignore_size_hints = TRUE;
+				gw->center = TRUE;
+				gw->recenter = TRUE;
+			} else if (is_wm_class (&hint, "gkrellm", 0)) {
+				/* hack, gkrell is stupid and doesn't set
+				 * right hints, such as USPosition and other
+				 * such stuff */
+				gw->center = FALSE;
+				gw->recenter = FALSE;
+			}
+			if (hint.res_name != NULL)
+				XFree (hint.res_name);
+			if (hint.res_class != NULL)
+				XFree (hint.res_class);
 		}
 
 		XGetGeometry (wm_disp, w,
@@ -445,7 +683,7 @@ add_window (Window w)
 		gw->x = x;
 		gw->y = x;
 
-		center_x_window (w, w);
+		center_x_window (gw, w, w);
 		add_deco (gw);
 
 		XAddToSaveSet (wm_disp, w);
@@ -471,8 +709,10 @@ remove_window (Window w)
 
 		gw->win = None;
 
-		XDestroyWindow (wm_disp, gw->deco);
-		gw->deco = None;
+		if (gw->deco != None) {
+			XDestroyWindow (wm_disp, gw->deco);
+			gw->deco = None;
+		}
 		if (gw->shadow != None) {
 			XDestroyWindow (wm_disp, gw->shadow);
 			gw->shadow = None;
@@ -524,7 +764,7 @@ add_all_current_windows (void)
 
 			if ( ! attribs.override_redirect &&
 			    attribs.map_state != IsUnmapped) {
-				add_window (children[i]);
+				add_window (children[i], FALSE /*center*/);
 			}
 		}
 
@@ -540,12 +780,15 @@ add_all_current_windows (void)
 static void
 reparent_to_root (GdmWindow *gw)
 {
-	trap_push ();
+	/* only if reparented */
+	if (gw->deco != None) {
+		trap_push ();
 
-	XReparentWindow (wm_disp, gw->win, wm_root, gw->x, gw->y);
-	XSync (wm_disp, False);
+		XReparentWindow (wm_disp, gw->win, wm_root, gw->x, gw->y);
+		XSync (wm_disp, False);
 
-	trap_pop ();
+		trap_pop ();
+	}
 }
 
 static void
@@ -587,7 +830,7 @@ event_process (XEvent *ev)
 		if (gw == NULL) {
 			if (ev->xmaprequest.parent == wm_root) {
 				XGrabServer (wm_disp);
-				add_window (w);
+				gw = add_window (w, TRUE /*center*/);
 				XUngrabServer (wm_disp);
 			}
 		}
@@ -600,7 +843,8 @@ event_process (XEvent *ev)
 		wchanges.border_width = ev->xconfigurerequest.border_width;
 		wchanges.sibling = ev->xconfigurerequest.above;
 		wchanges.stack_mode = ev->xconfigurerequest.detail;
-		if (gw == NULL) {
+		if (gw == NULL ||
+		    gw->deco == None) {
 			wchanges.x = ev->xconfigurerequest.x;
 			wchanges.y = ev->xconfigurerequest.y;
 		} else {
@@ -613,23 +857,25 @@ event_process (XEvent *ev)
 				  w,
 				  ev->xconfigurerequest.value_mask,
 				  &wchanges);
-		if (gw == NULL) {
-			center_x_window (w, w);
-		} else {
+		if (gw != NULL) {
 			gw->x = ev->xconfigurerequest.x;
 			gw->y = ev->xconfigurerequest.y;
-			wchanges.x = ev->xconfigurerequest.x - 1;
-			wchanges.y = ev->xconfigurerequest.y - 1;
-			wchanges.width = ev->xconfigurerequest.width + 2
-				+ 2*ev->xconfigurerequest.border_width;;
-			wchanges.height = ev->xconfigurerequest.height + 2
-				+ 2*ev->xconfigurerequest.border_width;;
-			wchanges.border_width = 0;
-			XConfigureWindow (wm_disp,
-					  gw->deco,
-					  ev->xconfigurerequest.value_mask,
-					  &wchanges);
-			center_x_window (gw->deco, gw->win);
+			if (gw->deco != None) {
+				wchanges.x = ev->xconfigurerequest.x - 1;
+				wchanges.y = ev->xconfigurerequest.y - 1;
+				wchanges.width = ev->xconfigurerequest.width + 2
+					+ 2*ev->xconfigurerequest.border_width;;
+				wchanges.height = ev->xconfigurerequest.height + 2
+					+ 2*ev->xconfigurerequest.border_width;;
+				wchanges.border_width = 0;
+				XConfigureWindow (wm_disp,
+						  gw->deco,
+						  ev->xconfigurerequest.value_mask,
+						  &wchanges);
+				center_x_window (gw, gw->deco, gw->win);
+			} else {
+				center_x_window (gw, gw->win, gw->win);
+			}
 			shadow_follow (gw);
 		}
 		XUngrabServer (wm_disp);
@@ -646,9 +892,15 @@ event_process (XEvent *ev)
 			if (ev->xcirculaterequest.place == PlaceOnTop) {
 				if (gw->shadow != None)
 					XRaiseWindow (wm_disp, gw->shadow);
-				XRaiseWindow (wm_disp, gw->deco);
+				if (gw->deco != None)
+					XRaiseWindow (wm_disp, gw->deco);
+				else
+					XRaiseWindow (wm_disp, gw->win);
 			} else {
-				XLowerWindow (wm_disp, gw->deco);
+				if (gw->deco != None)
+					XLowerWindow (wm_disp, gw->deco);
+				else
+					XLowerWindow (wm_disp, gw->win);
 				if (gw->shadow != None)
 					XLowerWindow (wm_disp, gw->shadow);
 			}
@@ -656,9 +908,10 @@ event_process (XEvent *ev)
 		break;
 	case MapNotify:
 		w = ev->xmap.window;
-		if ( ! ev->xmap.override_redirect) {
+		if ( ! ev->xmap.override_redirect &&
+		    focus_new_windows) {
 			gw = find_window (w, FALSE);
-			if (gw != NULL && focus_new_windows)
+			if (gw != NULL)
 				gdm_wm_focus_window (w);
 		}
 		break;
@@ -667,7 +920,8 @@ event_process (XEvent *ev)
 		gw = find_window (w, FALSE);
 		if (gw != NULL) {
 			XGrabServer (wm_disp);
-			XUnmapWindow (wm_disp, gw->deco);
+			if (gw->deco != None)
+				XUnmapWindow (wm_disp, gw->deco);
 			if (gw->shadow != None)
 				XUnmapWindow (wm_disp, gw->shadow);
 			reparent_to_root (gw);
@@ -754,6 +1008,28 @@ static GSourceFuncs event_funcs = {
 	(GDestroyNotify)g_free
 };
 
+/* HAAAAAAAAAAAAAACK */
+void
+gnome_dock_item_grab_pointer (GnomeDockItem *item)
+{
+  GdkCursor *fleur;
+
+  fleur = gdk_cursor_new (GDK_FLEUR);
+
+  /* Hm, not sure this is the right thing to do, but it seems to work.  */
+  while (gdk_pointer_grab (item->bin_window,
+                           FALSE,
+                           (GDK_BUTTON1_MOTION_MASK |
+                            GDK_POINTER_MOTION_HINT_MASK |
+                            GDK_BUTTON_RELEASE_MASK),
+                           NULL,
+                           fleur,
+                           GDK_CURRENT_TIME) != 0)
+	  gtk_main_iteration ();
+
+  gdk_cursor_destroy (fleur);
+}
+
 void
 gdm_wm_init (Window login_window)
 {
@@ -841,11 +1117,17 @@ gdm_wm_get_window_pos (Window window, int *xp, int *yp)
 		return;
 	}
 
-	XGetGeometry (wm_disp, gw->deco,
-		      &root, &x, &y, &width, &height, &border, &depth);
-
-	*xp = x + 1;
-	*yp = y + 1;
+	if (gw->deco != None) {
+		XGetGeometry (wm_disp, gw->deco,
+			      &root, &x, &y, &width, &height, &border, &depth);
+		*xp = x + 1;
+		*yp = y + 1;
+	} else {
+		XGetGeometry (wm_disp, gw->win,
+			      &root, &x, &y, &width, &height, &border, &depth);
+		*xp = x;
+		*yp = y;
+	}
 
 	trap_pop ();
 }
@@ -866,7 +1148,10 @@ gdm_wm_move_window_now (Window window, int x, int y)
 		trap_pop ();
 	}
 
-	XMoveWindow (wm_disp, gw->deco, x - 1, y - 1);
+	if (gw->deco != None)
+		XMoveWindow (wm_disp, gw->deco, x - 1, y - 1);
+	else
+		XMoveWindow (wm_disp, gw->win, x, y);
 	if (gw->shadow != None)
 		XMoveWindow (wm_disp, gw->deco, x + 4, y + 4);
 
